/* ------------------------- linea.c ---------------------------------------- */

/*
 * Draw lines on SVGA. Can be cery fast when usage of low level assembly
 * support is requested. It handles VGA page crossing.
 *
 * It originated from the DJgpp package but completely re-written since, the
 * only surviving clue being the WriteMode specification in the color field.
 * Actually this is not used either.
 *
 * These functions use some asm support from subs.asm.
 * (use -DUSE_ASM to include them):
 *
 * GrMoveTo (x, y)
 * GrDrawTo (x, y, color)
 * GrLine (x0, y0, x1, y1, color)
 *
 * These functions are pure asm and are found in subs.asm:
 *
 * GrMoveToA (x, y)
 * GrDrawToA (x, y, color)
 * GrLineA (x0, y0, x1, y1, color)
 *
 * If you only use the low-level asm version then this module is redundant.
*/

#include <stdlib.h>

#include "grh.h"

#define FAST_PLOT(va) \
	if (!_GrWriteMode)			\
		*va = (uchar)c;			\
	else if (_GrWriteMode & GrXOR)		\
		*va ^= c;			\
	else /* if (_GrWriteMode & GrXOR) */	\
		*va |= c


/* This does asm assisted drawing.
*/

extern int far _ILoop (uint iva, int dy, int dx, int dvx, int dvy, int c);
#define _InnerLoop	_ILoop

extern int far _SLoop (uint va, int count, int sv, int c);
#define _SimpleLoop	_SLoop


static uint	x1 = 0, y1 = 0, xydone = 0;
static uchar	far *va = VGA_PAGE;

extern void far
GrMoveTo (x, y)
uint	x;
uint	y;
{
#if 0
	if (x >= _GrSizeX)
		x = _GrSizeX-1;
	if (y >= _GrSizeY)
		y = _GrSizeY-1;
#endif
	x1 = x;
	y1 = y;
	xydone = 0;
}

extern void far
GrDrawTo (x2, y2, c)
uint	x2;
uint	y2;
uint	c;
{
	register int	dx, dy, svx, svy;
	int		t;

#if 0
	if (x2 >= _GrSizeX)
		x2 = _GrSizeX-1;
	if (y2 >= _GrSizeY)
		y2 = _GrSizeY-1;
#endif
	t = 0;
	if (!xydone) {
		t = 1;
		xydone = 1;
		MY_OFF(va) = (uint)_GrAddr (x1, y1);
		FAST_PLOT(va);
	}

	svx = 1;
	if ((dx = x2 - x1) < 0) {
		dx = -dx;
		svx = -svx;
	}

	svy = _GrSizeX;
	if ((dy = y2 - y1) < 0) {
		dy = -dy;
		svy = -svy;
	}

	if (dx > dy) {
		if (0 == dy) {
			uint	vb;

			vb = MY_OFF(va) + x2 - x1;	/* end point */

			if (!((int)(MY_OFF(va)^vb) < 0 &&
					(vb > 0xc000U || vb < 0x4000U))
#ifdef USE_ASM
			) {
				vgaset ((svx>0) ? MY_OFF(va)+1 : vb,
					c|_GrWriteMode, dx);
#else
			&& !_GrWriteMode
			) {
				memset (VGA_PAGE + (svx>0 ? MY_OFF(va)+1 : vb),
					c, dx);
#endif
				MY_OFF(va) = vb;
			} else
				MY_OFF(va) = _SimpleLoop (MY_OFF(va), dx, svx,
						c);
		} else
			MY_OFF(va) = _InnerLoop (MY_OFF(va), dy, dx, svx, svy,
					c);
		++GrStats[dx+t];
	} else if (dx < dy) {
		if (0 == dx)
			MY_OFF(va) = _SimpleLoop (MY_OFF(va), dy, svy, c);
		else
			MY_OFF(va) = _InnerLoop (MY_OFF(va), dx, dy, svy, svx,
					c);
		++GrStats[dy+t];
	} else /* if (dx == dy) */ {
		if (dx)
			MY_OFF(va) = _SimpleLoop (MY_OFF(va), dx, svx+svy, c);
		++GrStats[dx+t];
	}

	x1 = x2;
	y1 = y2;
}

extern void far
GrLine (xa, ya, xb, yb, c)
uint	xa;
uint	ya;
uint	xb;
uint	yb;
uint	c;
{
	if (xa != x1 || ya != y1 || !xydone)
		GrMoveTo (xa, ya);
#if 0
	if ((c&0xff00) != _GrWriteMode)
		GrSetWriteMode (c);
#endif
	GrDrawTo (xb, yb, c);
}
